/*
 * Copyright (c) 2007-2011, Lloyd Hilaiel <lloyd@hilaiel.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef __YAJL_LEX_H__
#define __YAJL_LEX_H__

#include "yajl_common.h"

typedef enum {
    bson_yajl_tok_bool,         
    bson_yajl_tok_colon,
    bson_yajl_tok_comma,     
    bson_yajl_tok_eof,
    bson_yajl_tok_error,
    bson_yajl_tok_left_brace,     
    bson_yajl_tok_left_bracket,
    bson_yajl_tok_null,         
    bson_yajl_tok_right_brace,     
    bson_yajl_tok_right_bracket,

    /* we differentiate between integers and doubles to allow the
     * parser to interpret the number without re-scanning */
    bson_yajl_tok_integer, 
    bson_yajl_tok_double, 

    /* we differentiate between strings which require further processing,
     * and strings that do not */
    bson_yajl_tok_string,
    bson_yajl_tok_string_with_escapes,

    /* comment tokens are not currently returned to the parser, ever */
    bson_yajl_tok_comment
} bson_yajl_tok;

typedef struct bson_yajl_lexer_t * bson_yajl_lexer;

bson_yajl_lexer bson_yajl_lex_alloc(bson_yajl_alloc_funcs * alloc,
                          unsigned int allowComments,
                          unsigned int validateUTF8);

void bson_yajl_lex_free(bson_yajl_lexer lexer);

/**
 * run/continue a lex. "offset" is an input/output parameter.
 * It should be initialized to zero for a
 * new chunk of target text, and upon subsetquent calls with the same
 * target text should passed with the value of the previous invocation.
 *
 * the client may be interested in the value of offset when an error is
 * returned from the lexer.  This allows the client to render useful
n * error messages.
 *
 * When you pass the next chunk of data, context should be reinitialized
 * to zero.
 * 
 * Finally, the output buffer is usually just a pointer into the jsonText,
 * however in cases where the entity being lexed spans multiple chunks,
 * the lexer will buffer the entity and the data returned will be
 * a pointer into that buffer.
 *
 * This behavior is abstracted from client code except for the performance
 * implications which require that the client choose a reasonable chunk
 * size to get adequate performance.
 */
bson_yajl_tok bson_yajl_lex_lex(bson_yajl_lexer lexer, const unsigned char * jsonText,
                      size_t jsonTextLen, size_t * offset,
                      const unsigned char ** outBuf, size_t * outLen);

/** have a peek at the next token, but don't move the lexer forward */
bson_yajl_tok bson_yajl_lex_peek(bson_yajl_lexer lexer, const unsigned char * jsonText,
                       size_t jsonTextLen, size_t offset);


typedef enum {
    bson_yajl_lex_e_ok = 0,
    bson_yajl_lex_string_invalid_utf8,
    bson_yajl_lex_string_invalid_escaped_char,
    bson_yajl_lex_string_invalid_json_char,
    bson_yajl_lex_string_invalid_hex_char,
    bson_yajl_lex_invalid_char,
    bson_yajl_lex_invalid_string,
    bson_yajl_lex_missing_integer_after_decimal,
    bson_yajl_lex_missing_integer_after_exponent,
    bson_yajl_lex_missing_integer_after_minus,
    bson_yajl_lex_unallowed_comment
} bson_yajl_lex_error;

const char * bson_yajl_lex_error_to_string(bson_yajl_lex_error error);

/** allows access to more specific information about the lexical
 *  error when bson_yajl_lex_lex returns bson_yajl_tok_error. */
bson_yajl_lex_error bson_yajl_lex_get_error(bson_yajl_lexer lexer);

/** get the current offset into the most recently lexed json string. */
size_t bson_yajl_lex_current_offset(bson_yajl_lexer lexer);

/** get the number of lines lexed by this lexer instance */
size_t bson_yajl_lex_current_line(bson_yajl_lexer lexer);

/** get the number of chars lexed by this lexer instance since the last
 *  \n or \r */
size_t bson_yajl_lex_current_char(bson_yajl_lexer lexer);

#endif
