/*
 * Copyright (c) 2007-2011, Lloyd Hilaiel <lloyd@hilaiel.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "yajl_gen.h"
#include "yajl_buf.h"
#include "yajl_encode.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <stdarg.h>

typedef enum {
    bson_yajl_gen_start,
    bson_yajl_gen_map_start,
    bson_yajl_gen_map_key,
    bson_yajl_gen_map_val,
    bson_yajl_gen_array_start,
    bson_yajl_gen_in_array,
    bson_yajl_gen_complete,
    bson_yajl_gen_error
} bson_yajl_gen_state;

struct bson_yajl_gen_t
{
    unsigned int flags;
    unsigned int depth;
    const char * indentString;
    bson_yajl_gen_state state[YAJL_MAX_DEPTH];
    bson_yajl_print_t print;
    void * ctx; /* bson_yajl_buf */
    /* memory allocation routines */
    bson_yajl_alloc_funcs alloc;
};

int
bson_yajl_gen_config(bson_yajl_gen g, bson_yajl_gen_option opt, ...)
{
    int rv = 1;
    va_list ap;
    va_start(ap, opt);

    switch(opt) {
        case bson_yajl_gen_beautify:
        case bson_yajl_gen_validate_utf8:
        case bson_yajl_gen_escape_solidus:
            if (va_arg(ap, int)) g->flags |= opt;
            else g->flags &= ~opt;
            break;
        case bson_yajl_gen_indent_string: {
            const char *indent = va_arg(ap, const char *);
            g->indentString = indent;
            for (; *indent; indent++) {
                if (*indent != '\n'
                    && *indent != '\v'
                    && *indent != '\f'
                    && *indent != '\t'
                    && *indent != '\r'
                    && *indent != ' ')
                {
                    g->indentString = NULL;
                    rv = 0;
                }
            }
            break;
        }
        case bson_yajl_gen_print_callback:
            bson_yajl_buf_free(g->ctx);
            g->print = va_arg(ap, const bson_yajl_print_t);
            g->ctx = va_arg(ap, void *);
            break;
        default:
            rv = 0;
    }

    va_end(ap);

    return rv;
}



bson_yajl_gen
bson_yajl_gen_alloc(const bson_yajl_alloc_funcs * afs)
{
    bson_yajl_gen g = NULL;
    bson_yajl_alloc_funcs afsBuffer;

    /* first order of business is to set up memory allocation routines */
    if (afs != NULL) {
        if (afs->malloc == NULL || afs->realloc == NULL || afs->free == NULL)
        {
            return NULL;
        }
    } else {
        bson_yajl_set_default_alloc_funcs(&afsBuffer);
        afs = &afsBuffer;
    }

    g = (bson_yajl_gen) YA_MALLOC(afs, sizeof(struct bson_yajl_gen_t));
    if (!g) return NULL;

    memset((void *) g, 0, sizeof(struct bson_yajl_gen_t));
    /* copy in pointers to allocation routines */
    memcpy((void *) &(g->alloc), (void *) afs, sizeof(bson_yajl_alloc_funcs));

    g->print = (bson_yajl_print_t)&bson_yajl_buf_append;
    g->ctx = bson_yajl_buf_alloc(&(g->alloc));
    g->indentString = "    ";

    return g;
}

void
bson_yajl_gen_free(bson_yajl_gen g)
{
    if (g->print == (bson_yajl_print_t)&bson_yajl_buf_append) bson_yajl_buf_free((bson_yajl_buf)g->ctx);
    YA_FREE(&(g->alloc), g);
}

#define INSERT_SEP \
    if (g->state[g->depth] == bson_yajl_gen_map_key ||               \
        g->state[g->depth] == bson_yajl_gen_in_array) {              \
        g->print(g->ctx, ",", 1);                               \
        if ((g->flags & bson_yajl_gen_beautify)) g->print(g->ctx, "\n", 1);               \
    } else if (g->state[g->depth] == bson_yajl_gen_map_val) {        \
        g->print(g->ctx, ":", 1);                               \
        if ((g->flags & bson_yajl_gen_beautify)) g->print(g->ctx, " ", 1);                \
   }

#define INSERT_WHITESPACE                                               \
    if ((g->flags & bson_yajl_gen_beautify)) {                                                    \
        if (g->state[g->depth] != bson_yajl_gen_map_val) {                   \
            unsigned int _i;                                            \
            for (_i=0;_i<g->depth;_i++)                                 \
                g->print(g->ctx,                                        \
                         g->indentString,                               \
                         (unsigned int)strlen(g->indentString));        \
        }                                                               \
    }

#define ENSURE_NOT_KEY \
    if (g->state[g->depth] == bson_yajl_gen_map_key ||       \
        g->state[g->depth] == bson_yajl_gen_map_start)  {    \
        return bson_yajl_gen_keys_must_be_strings;           \
    }                                                   \

/* check that we're not complete, or in error state.  in a valid state
 * to be generating */
#define ENSURE_VALID_STATE \
    if (g->state[g->depth] == bson_yajl_gen_error) {   \
        return bson_yajl_gen_in_error_state;\
    } else if (g->state[g->depth] == bson_yajl_gen_complete) {   \
        return bson_yajl_gen_generation_complete;                \
    }

#define INCREMENT_DEPTH \
    if (++(g->depth) >= YAJL_MAX_DEPTH) return bson_yajl_max_depth_exceeded;

#define DECREMENT_DEPTH \
  if (--(g->depth) >= YAJL_MAX_DEPTH) return bson_yajl_gen_invalid_string;

#define APPENDED_ATOM \
    switch (g->state[g->depth]) {                   \
        case bson_yajl_gen_start:                        \
            g->state[g->depth] = bson_yajl_gen_complete; \
            break;                                  \
        case bson_yajl_gen_map_start:                    \
        case bson_yajl_gen_map_key:                      \
            g->state[g->depth] = bson_yajl_gen_map_val;  \
            break;                                  \
        case bson_yajl_gen_array_start:                  \
            g->state[g->depth] = bson_yajl_gen_in_array; \
            break;                                  \
        case bson_yajl_gen_map_val:                      \
            g->state[g->depth] = bson_yajl_gen_map_key;  \
            break;                                  \
        default:                                    \
            break;                                  \
    }                                               \

#define FINAL_NEWLINE                                        \
    if ((g->flags & bson_yajl_gen_beautify) && g->state[g->depth] == bson_yajl_gen_complete) \
        g->print(g->ctx, "\n", 1);

bson_yajl_gen_status
bson_yajl_gen_integer(bson_yajl_gen g, long long int number)
{
    char i[32];
    ENSURE_VALID_STATE; ENSURE_NOT_KEY; INSERT_SEP; INSERT_WHITESPACE;
    sprintf(i, "%lld", number);
    g->print(g->ctx, i, (unsigned int)strlen(i));
    APPENDED_ATOM;
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

#if defined(_WIN32) || defined(WIN32)
# include <float.h>
# ifndef isnan
#  define isnan _isnan
# endif
# ifndef isinf
#  define isinf !_finite
# endif
#endif

#if defined(__sun) && defined(__GNUC__) && __GNUC__ < 4
# define isinf !finite
#endif

bson_yajl_gen_status
bson_yajl_gen_double(bson_yajl_gen g, double number)
{
    char i[32];
    ENSURE_VALID_STATE; ENSURE_NOT_KEY;
    if (isnan(number) || isinf(number)) return bson_yajl_gen_invalid_number;
    INSERT_SEP; INSERT_WHITESPACE;
    sprintf(i, "%.20g", number);
    if (strspn(i, "0123456789-") == strlen(i)) {
        strcat(i, ".0");
    }
    g->print(g->ctx, i, (unsigned int)strlen(i));
    APPENDED_ATOM;
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_number(bson_yajl_gen g, const char * s, size_t l)
{
    ENSURE_VALID_STATE; ENSURE_NOT_KEY; INSERT_SEP; INSERT_WHITESPACE;
    g->print(g->ctx, s, l);
    APPENDED_ATOM;
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_string(bson_yajl_gen g, const unsigned char * str,
                size_t len)
{
    // if validation is enabled, check that the string is valid utf8
    // XXX: This checking could be done a little faster, in the same pass as
    // the string encoding
    if (g->flags & bson_yajl_gen_validate_utf8) {
        if (!bson_yajl_string_validate_utf8(str, len)) {
            return bson_yajl_gen_invalid_string;
        }
    }
    ENSURE_VALID_STATE; INSERT_SEP; INSERT_WHITESPACE;
    g->print(g->ctx, "\"", 1);
    bson_yajl_string_encode(g->print, g->ctx, str, len, g->flags & bson_yajl_gen_escape_solidus);
    g->print(g->ctx, "\"", 1);
    APPENDED_ATOM;
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_null(bson_yajl_gen g)
{
    ENSURE_VALID_STATE; ENSURE_NOT_KEY; INSERT_SEP; INSERT_WHITESPACE;
    g->print(g->ctx, "null", strlen("null"));
    APPENDED_ATOM;
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_bool(bson_yajl_gen g, int boolean)
{
    const char * val = boolean ? "true" : "false";

	ENSURE_VALID_STATE; ENSURE_NOT_KEY; INSERT_SEP; INSERT_WHITESPACE;
    g->print(g->ctx, val, (unsigned int)strlen(val));
    APPENDED_ATOM;
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_map_open(bson_yajl_gen g)
{
    ENSURE_VALID_STATE; ENSURE_NOT_KEY; INSERT_SEP; INSERT_WHITESPACE;
    INCREMENT_DEPTH;

    g->state[g->depth] = bson_yajl_gen_map_start;
    g->print(g->ctx, "{", 1);
    if ((g->flags & bson_yajl_gen_beautify)) g->print(g->ctx, "\n", 1);
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_map_close(bson_yajl_gen g)
{
    ENSURE_VALID_STATE;
    DECREMENT_DEPTH;

    if ((g->flags & bson_yajl_gen_beautify)) g->print(g->ctx, "\n", 1);
    APPENDED_ATOM;
    INSERT_WHITESPACE;
    g->print(g->ctx, "}", 1);
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_array_open(bson_yajl_gen g)
{
    ENSURE_VALID_STATE; ENSURE_NOT_KEY; INSERT_SEP; INSERT_WHITESPACE;
    INCREMENT_DEPTH;
    g->state[g->depth] = bson_yajl_gen_array_start;
    g->print(g->ctx, "[", 1);
    if ((g->flags & bson_yajl_gen_beautify)) g->print(g->ctx, "\n", 1);
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_array_close(bson_yajl_gen g)
{
    ENSURE_VALID_STATE;
    DECREMENT_DEPTH;
    if ((g->flags & bson_yajl_gen_beautify)) g->print(g->ctx, "\n", 1);
    APPENDED_ATOM;
    INSERT_WHITESPACE;
    g->print(g->ctx, "]", 1);
    FINAL_NEWLINE;
    return bson_yajl_gen_status_ok;
}

bson_yajl_gen_status
bson_yajl_gen_get_buf(bson_yajl_gen g, const unsigned char ** buf,
                 size_t * len)
{
    if (g->print != (bson_yajl_print_t)&bson_yajl_buf_append) return bson_yajl_gen_no_buf;
    *buf = bson_yajl_buf_data((bson_yajl_buf)g->ctx);
    *len = bson_yajl_buf_len((bson_yajl_buf)g->ctx);
    return bson_yajl_gen_status_ok;
}

void
bson_yajl_gen_clear(bson_yajl_gen g)
{
    if (g->print == (bson_yajl_print_t)&bson_yajl_buf_append) bson_yajl_buf_clear((bson_yajl_buf)g->ctx);
}
