/*
 * Copyright (c) 2007-2011, Lloyd Hilaiel <lloyd@hilaiel.com>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * \file bson_yajl_gen.h
 * Interface to YAJL's JSON generation facilities.
 */

#include <yajl/yajl_common.h>

#ifndef __YAJL_GEN_H__
#define __YAJL_GEN_H__

#include <stddef.h>

#ifdef __cplusplus
extern "C" {
#endif
    /** generator status codes */
    typedef enum {
        /** no error */
        bson_yajl_gen_status_ok = 0,
        /** at a point where a map key is generated, a function other than
         *  bson_yajl_gen_string was called */
        bson_yajl_gen_keys_must_be_strings,
        /** YAJL's maximum generation depth was exceeded.  see
         *  YAJL_MAX_DEPTH */
        bson_yajl_max_depth_exceeded,
        /** A generator function (bson_yajl_gen_XXX) was called while in an error
         *  state */
        bson_yajl_gen_in_error_state,
        /** A complete JSON document has been generated */
        bson_yajl_gen_generation_complete,                
        /** bson_yajl_gen_double was passed an invalid floating point value
         *  (infinity or NaN). */
        bson_yajl_gen_invalid_number,
        /** A print callback was passed in, so there is no internal
         * buffer to get from */
        bson_yajl_gen_no_buf,
        /** returned from bson_yajl_gen_string() when the bson_yajl_gen_validate_utf8
         *  option is enabled and an invalid was passed by client code.
         */
        bson_yajl_gen_invalid_string
    } bson_yajl_gen_status;

    /** an opaque handle to a generator */
    typedef struct bson_yajl_gen_t * bson_yajl_gen;

    /** a callback used for "printing" the results. */
    typedef void (*bson_yajl_print_t)(void * ctx,
                                 const char * str,
                                 size_t len);

    /** configuration parameters for the parser, these may be passed to
     *  bson_yajl_gen_config() along with option specific argument(s).  In general,
     *  all configuration parameters default to *off*. */
    typedef enum {
        /** generate indented (beautiful) output */
        bson_yajl_gen_beautify = 0x01,
        /**
         * Set an indent string which is used when bson_yajl_gen_beautify
         * is enabled.  Maybe something like \\t or some number of
         * spaces.  The default is four spaces ' '.
         */
        bson_yajl_gen_indent_string = 0x02,
        /**
         * Set a function and context argument that should be used to
         * output generated json.  the function should conform to the
         * bson_yajl_print_t prototype while the context argument is a
         * void * of your choosing.
         *
         * example:
         *   bson_yajl_gen_config(g, bson_yajl_gen_print_callback, myFunc, myVoidPtr);
         */
        bson_yajl_gen_print_callback = 0x04,
        /**
         * Normally the generator does not validate that strings you
         * pass to it via bson_yajl_gen_string() are valid UTF8.  Enabling
         * this option will cause it to do so.
         */
        bson_yajl_gen_validate_utf8 = 0x08,
        /**
         * the forward solidus (slash or '/' in human) is not required to be
         * escaped in json text.  By default, YAJL will not escape it in the
         * iterest of saving bytes.  Setting this flag will cause YAJL to
         * always escape '/' in generated JSON strings.
         */
        bson_yajl_gen_escape_solidus = 0x10
    } bson_yajl_gen_option;

    /** allow the modification of generator options subsequent to handle
     *  allocation (via bson_yajl_alloc)
     *  \returns zero in case of errors, non-zero otherwise
     */
    YAJL_API int bson_yajl_gen_config(bson_yajl_gen g, bson_yajl_gen_option opt, ...);

    /** allocate a generator handle
     *  \param allocFuncs an optional pointer to a structure which allows
     *                    the client to overide the memory allocation
     *                    used by yajl.  May be NULL, in which case
     *                    malloc/free/realloc will be used.
     *
     *  \returns an allocated handle on success, NULL on failure (bad params)
     */
    YAJL_API bson_yajl_gen bson_yajl_gen_alloc(const bson_yajl_alloc_funcs * allocFuncs);

    /** free a generator handle */
    YAJL_API void bson_yajl_gen_free(bson_yajl_gen handle);

    YAJL_API bson_yajl_gen_status bson_yajl_gen_integer(bson_yajl_gen hand, long long int number);
    /** generate a floating point number.  number may not be infinity or
     *  NaN, as these have no representation in JSON.  In these cases the
     *  generator will return 'bson_yajl_gen_invalid_number' */
    YAJL_API bson_yajl_gen_status bson_yajl_gen_double(bson_yajl_gen hand, double number);
    YAJL_API bson_yajl_gen_status bson_yajl_gen_number(bson_yajl_gen hand,
                                             const char * num,
                                             size_t len);
    YAJL_API bson_yajl_gen_status bson_yajl_gen_string(bson_yajl_gen hand,
                                             const unsigned char * str,
                                             size_t len);
    YAJL_API bson_yajl_gen_status bson_yajl_gen_null(bson_yajl_gen hand);
    YAJL_API bson_yajl_gen_status bson_yajl_gen_bool(bson_yajl_gen hand, int boolean);
    YAJL_API bson_yajl_gen_status bson_yajl_gen_map_open(bson_yajl_gen hand);
    YAJL_API bson_yajl_gen_status bson_yajl_gen_map_close(bson_yajl_gen hand);
    YAJL_API bson_yajl_gen_status bson_yajl_gen_array_open(bson_yajl_gen hand);
    YAJL_API bson_yajl_gen_status bson_yajl_gen_array_close(bson_yajl_gen hand);

    /** access the null terminated generator buffer.  If incrementally
     *  outputing JSON, one should call bson_yajl_gen_clear to clear the
     *  buffer.  This allows stream generation. */
    YAJL_API bson_yajl_gen_status bson_yajl_gen_get_buf(bson_yajl_gen hand,
                                              const unsigned char ** buf,
                                              size_t * len);

    /** clear yajl's output buffer, but maintain all internal generation
     *  state.  This function will not "reset" the generator state, and is
     *  intended to enable incremental JSON outputing. */
    YAJL_API void bson_yajl_gen_clear(bson_yajl_gen hand);

#ifdef __cplusplus
}
#endif    

#endif
